# -*- coding: utf-8 -*-
"""Contains the main `APISpec` class.
"""
from .core import APISpec, Path

__version__ = '0.17.3'
__author__ = 'Steven Loria, Josh Carp, and contributors'
__license__ = 'MIT'


__all__ = [
    'APISpec',
    'Path',
]
